package com.scribetechglobal.augnito_flutter_sdk

import android.Manifest
import android.app.Activity
import android.content.Context
import android.content.pm.PackageManager
import android.media.AudioFormat
import android.media.AudioRecord
import android.media.AudioRecord.OnRecordPositionUpdateListener
import android.media.MediaRecorder
import android.util.Log
import androidx.annotation.NonNull
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry
import java.nio.ByteBuffer
import java.nio.ByteOrder

enum class SoundStreamErrors {
    FailedToRecord,
    PermissionsMissing,
    Unknown,
}

enum class SoundStreamStatus {
    Unset,
    Initialized,
    Recording,
    Stopped,
}

const val methodChannelName = "com.scribetechglobal.augnito_flutter_sdk/methods"

/** AugnitoFlutterSdkPlugin */
class AugnitoFlutterSdkPlugin : FlutterPlugin, MethodCallHandler,
    PluginRegistry.RequestPermissionsResultListener,
    ActivityAware {
    /// The MethodChannel that will the communication between Flutter and native Android
    ///
    /// This local reference serves to register the plugin with the Flutter Engine and unregister it
    /// when the Flutter Engine is detached from the Activity
    private lateinit var channel: MethodChannel

    private val tag = "AugnitoMicStream"
    private val audioRecordPermissionCode = 14887
    private var loggingEnabled: Boolean = false
    private var activeResult: Result? = null
    private var pluginContext: Context? = null
    private var currentActivity: Activity? = null
    private var permissionToRecordAudio: Boolean = false

    // #region recording vars
    private val recordFormat = AudioFormat.ENCODING_PCM_16BIT
    private var sampleRate = 16000 // 16Khz
    private var audioData: ShortArray? = null
    private var recorderBufferSize = 8192
    private var periodFrames = 8192
    private var recorder: AudioRecord? = null
    private var listener: OnRecordPositionUpdateListener? = null
    // #endregion

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        pluginContext = flutterPluginBinding.applicationContext
        channel = MethodChannel(flutterPluginBinding.binaryMessenger, methodChannelName)
        channel.setMethodCallHandler(this)
    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        channel.setMethodCallHandler(null)
    }

    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        when (call.method) {
            "initializeRecorder" -> initializeRecorder(call, result)
            "startRecording" -> startRecording(result)
            "stopRecording" -> stopRecording(result)
            "checkPermission" -> hasPermission(result)
            "dispose" -> dispose(result)
            else -> result.notImplemented()
        }
    }

    override fun onDetachedFromActivity() {
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        currentActivity = binding.activity
        binding.addRequestPermissionsResultListener(this)
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        currentActivity = binding.activity
        binding.addRequestPermissionsResultListener(this)
    }

    override fun onDetachedFromActivityForConfigChanges() {
    }

    // region private

    // region permissions
    private fun hasRecordPermission(): Boolean {
        if (permissionToRecordAudio) return true

        val localContext = pluginContext
        permissionToRecordAudio = localContext != null && ContextCompat.checkSelfPermission(
            localContext,
            Manifest.permission.RECORD_AUDIO
        ) == PackageManager.PERMISSION_GRANTED
        return permissionToRecordAudio
    }

    private fun hasPermission(result: Result) {
        result.success(hasRecordPermission())
    }

    private fun requestRecordPermission() {
        val localActivity = currentActivity
        if (!hasRecordPermission() && localActivity != null) {
            log("requesting RECORD_AUDIO permission")
            ActivityCompat.requestPermissions(
                localActivity,
                arrayOf(Manifest.permission.RECORD_AUDIO), audioRecordPermissionCode
            )
        }
    }

    override fun onRequestPermissionsResult(
        requestCode: Int,
        permissions: Array<out String>,
        grantResults: IntArray
    ): Boolean {
        when (requestCode) {
            audioRecordPermissionCode -> {
                if (grantResults != null) {
                    permissionToRecordAudio = grantResults.isNotEmpty() &&
                            grantResults[0] == PackageManager.PERMISSION_GRANTED
                }
                completeInitializeRecorder()
                return true
            }
        }
        return false
    }
    // endregion

    private fun initializeRecorder(@NonNull call: MethodCall, @NonNull result: Result) {
        // params
        sampleRate = call.argument<Int>("sampleRate") ?: sampleRate
        loggingEnabled = call.argument<Boolean>("loggingEnabled") ?: false

        periodFrames =
            AudioRecord.getMinBufferSize(sampleRate, AudioFormat.CHANNEL_IN_MONO, recordFormat)
        recorderBufferSize = periodFrames * 2
        audioData = ShortArray(periodFrames)
        activeResult = result

        val localContext = pluginContext
        if (null == localContext) {
            completeInitializeRecorder()
            return
        }
        permissionToRecordAudio = ContextCompat.checkSelfPermission(
            localContext,
            Manifest.permission.RECORD_AUDIO
        ) == PackageManager.PERMISSION_GRANTED
        if (!permissionToRecordAudio) {
            requestRecordPermission()
        } else {
            log("has permission, completing")
            completeInitializeRecorder()
        }
        log("leaving initializeIfPermitted")
    }

    private fun initRecorder() {
        if (recorder?.state == AudioRecord.STATE_INITIALIZED) {
            return
        }
        recorder = AudioRecord(
            MediaRecorder.AudioSource.DEFAULT,
            sampleRate,
            AudioFormat.CHANNEL_IN_MONO,
            recordFormat,
            recorderBufferSize
        )
        if (recorder != null) {
            listener = createRecordListener(recorderBufferSize, periodFrames)
            recorder?.positionNotificationPeriod = periodFrames
            recorder?.setRecordPositionUpdateListener(listener)
        }
    }

    private fun completeInitializeRecorder() {
        log("completeInitialize")
        val initResult: HashMap<String, Any> = HashMap()

        if (permissionToRecordAudio) {
            recorder?.release()
            initRecorder()
            initResult["isMeteringEnabled"] = true
            sendRecorderStatus(SoundStreamStatus.Initialized)
        }

        // return true if there was permission to record audio(recorded was created)
        log("sending result")
        activeResult?.success(permissionToRecordAudio)
        log("leaving complete")
        activeResult = null
    }

    private fun createRecordListener(
        bufferSize: Int,
        periodFrames: Int
    ): OnRecordPositionUpdateListener? {
        return object : OnRecordPositionUpdateListener {
            override fun onMarkerReached(recorder: AudioRecord) {
                recorder.read(audioData!!, 0, bufferSize)
            }

            override fun onPeriodicNotification(recorder: AudioRecord) {
                val data = audioData!!
                val shortOut = recorder.read(data, 0, periodFrames)
                // https://flutter.io/platform-channels/#codec
                // convert short to int because of platform-channel's limitation
                val byteBuffer = ByteBuffer.allocate(shortOut * 2)
                byteBuffer.order(ByteOrder.LITTLE_ENDIAN).asShortBuffer().put(data)

                sendEventMethod("dataPeriod", byteBuffer.array())
            }
        }
    }

    private fun startRecording(result: Result) {
        try {
            if (recorder!!.recordingState == AudioRecord.RECORDSTATE_RECORDING) {
                result.success(true)
                return
            }
            recorder!!.startRecording()
            sendRecorderStatus(SoundStreamStatus.Recording)
            result.success(true)
        } catch (e: IllegalStateException) {
            log("record() failed")
            result.error(
                SoundStreamErrors.FailedToRecord.name,
                "Failed to start recording",
                e.localizedMessage
            )
        }
    }

    private fun dispose(result: Result) {
        try {
            recorder?.release()
            recorder = null
        } catch (e: IllegalStateException) {
            log("dispose() failed")
            result.error(
                SoundStreamErrors.Unknown.name,
                "There was an error while disposing the recorder.",
                e.localizedMessage
            )
        }
    }

    private fun stopRecording(result: Result) {
        try {
            if (recorder!!.recordingState == AudioRecord.RECORDSTATE_STOPPED) {
                result.success(true)
                return
            }
            recorder!!.stop()
            sendRecorderStatus(SoundStreamStatus.Stopped)
            result.success(true)
        } catch (e: IllegalStateException) {
            log("record() failed")
            result.error(
                SoundStreamErrors.FailedToRecord.name,
                "Failed to start recording",
                e.localizedMessage
            )
        }
    }

    private fun log(message: String) {
        if (loggingEnabled) {
            Log.d(tag, message)
        }
    }

    private fun sendRecorderStatus(status: SoundStreamStatus) {
        sendEventMethod("recorderStatus", status.name)
    }

    private fun sendEventMethod(name: String, data: Any) {
        val eventData: HashMap<String, Any> = HashMap()
        eventData["name"] = name
        eventData["data"] = data
        channel.invokeMethod("platformEvent", eventData)
    }
    // endregion
}
