import Flutter
import UIKit
import AVFoundation

public enum SoundStreamErrors: String {
    case FailedToRecord
    case PermissionsMissing
    case Unknown
}

public enum SoundStreamStatus: String {
    case Unset
    case Initialized
    case Recording
    case Stopped
}

@available(iOS 9.0, *)
public class SwiftAugnitoFlutterSdkPlugin: NSObject, FlutterPlugin {
    // internal and support
    private var channel: FlutterMethodChannel
    static let methodChannelName:String = "com.scribetechglobal.augnito_flutter_sdk/methods"
    private var registrar: FlutterPluginRegistrar
    private var hasPermission: Bool = false
    private var debugLogging: Bool = false
    
    // record specific
    private var audioEngine:AVAudioEngine?
    private let recordBus = 0
    private var inputNode: AVAudioInputNode?
    private var recordSampleRate: Double = 16000 // 16Khz
    private var recordBufferSize: AVAudioFrameCount = 8192
    private var recordChannel = 0
    private var recordSettings: [String:Int]!
    private var recordFormat: AVAudioFormat!
    
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: methodChannelName, binaryMessenger: registrar.messenger())
        let instance = SwiftAugnitoFlutterSdkPlugin(channel, registrar: registrar)
        registrar.addMethodCallDelegate(instance, channel: channel)
    }
    
    init( _ channel: FlutterMethodChannel, registrar: FlutterPluginRegistrar ) {
        self.channel = channel
        self.registrar = registrar
        super.init()
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        case "initializeRecorder":
            attemptRecorderInitialization(call, result)
        case "startRecording":
            startRecording(result)
        case "stopRecording":
            stopRecording(result)
        case "checkPermission":
            checkPermission(result)
        case "dispose":
            break
        default:
            print("Unrecognized method: \(call.method)")
            sendResult(result, FlutterMethodNotImplemented)
        }
    }
    
    
    // #region private and plugin methods
    
    private func sendResult(_ result: @escaping FlutterResult, _ arguments: Any?) {
        DispatchQueue.main.async {
            result( arguments )
        }
    }
    
    private func invokeFlutter( _ method: String, _ arguments: Any? ) {
        DispatchQueue.main.async {
            self.channel.invokeMethod( method, arguments: arguments )
        }
    }
    
    private func checkAndRequestPermission(completion callback: @escaping ((Bool) -> Void)) {
        if (hasPermission) {
            callback(hasPermission)
            return
        }
        var permission: AVAudioSession.RecordPermission
#if swift(>=4.2)
        permission = AVAudioSession.sharedInstance().recordPermission
#else
        permission = AVAudioSession.sharedInstance().recordPermission()
#endif
        switch permission {
        case .granted:
            printMessage("granted")
            hasPermission = true
            callback(hasPermission)
            break
        case .denied:
            printMessage("denied")
            hasPermission = false
            callback(hasPermission)
            break
        case .undetermined:
            printMessage("undetermined")
            AVAudioSession.sharedInstance().requestRecordPermission() { [unowned self] allowed in
                if allowed {
                    self.hasPermission = true
                    printMessage("undetermined true")
                    callback(self.hasPermission)
                } else {
                    self.hasPermission = false
                    printMessage("undetermined false")
                    callback(self.hasPermission)
                }
            }
            break
        default:
            callback(hasPermission)
            break
        }
    }
    
    private func checkPermission(_ result: @escaping FlutterResult) {
        if (hasPermission) {
            result(hasPermission)
            return
        }
        var permission: AVAudioSession.RecordPermission
#if swift(>=4.2)
        permission = AVAudioSession.sharedInstance().recordPermission
#else
        permission = AVAudioSession.sharedInstance().recordPermission()
#endif
        switch permission {
        case .granted:
            printMessage("granted")
            hasPermission = true
            result(hasPermission)
            break
        case .denied:
            printMessage("denied")
            hasPermission = false
            result(hasPermission)
            break

        default:
            result(hasPermission)
            break
        }
    }
    
    private func hasPermission( _ result: @escaping FlutterResult) {
        checkAndRequestPermission { value in
            self.sendResult(result, value)
        }
    }
    
    private func startEngine() {
        guard !audioEngine!.isRunning else {
            return
        }
        
        try? audioEngine!.start()
    }
    
    private func stopEngine() {
        audioEngine!.stop()
        audioEngine!.reset()
    }
    
    private func sendEventMethod(_ name: String, _ data: Any) {
        var eventData: [String: Any] = [:]
        eventData["name"] = name
        eventData["data"] = data
        invokeFlutter("platformEvent", eventData)
    }
    
    private func sendRecorderStatus(_ status: SoundStreamStatus) {
        sendEventMethod("recorderStatus", status.rawValue)
    }
    
    // attempts to initialize the recorder. Fails if permissions are not granted
    private func attemptRecorderInitialization(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        checkAndRequestPermission { isGranted in
            if isGranted {
                self.initializeRecorder(call,  result)
            } else {
                self.sendResult(result, FlutterError( code: SoundStreamErrors.PermissionsMissing.rawValue,
                                                      message:"Unable to initialize: RECORD_AUDIO permission not granted.",
                                                      details: nil ))
            }
        }
    }
    
    private func initializeRecorder(_ call: FlutterMethodCall, _ result: @escaping FlutterResult) {
        guard let argsArr = call.arguments as? Dictionary<String,AnyObject>
        else {
            sendResult(result, FlutterError( code: SoundStreamErrors.Unknown.rawValue,
                                             message:"Incorrect parameters",
                                             details: nil ))
            return
        }
        
        if(audioEngine == nil){
            initializeAudioEngine()
        }
        
        recordSampleRate = argsArr["sampleRate"] as? Double ?? recordSampleRate
        debugLogging = argsArr["showLogs"] as? Bool ?? debugLogging
        recordFormat = AVAudioFormat(commonFormat: AVAudioCommonFormat.pcmFormatInt16, sampleRate: recordSampleRate, channels: 1, interleaved: true)
        sendRecorderStatus(SoundStreamStatus.Initialized)
        sendResult(result, true)
    }
    
    public func initializeAudioEngine() {
        self.audioEngine = AVAudioEngine()
        self.inputNode = audioEngine!.inputNode
        audioEngine!.prepare()
    }
    
    private func resetEngineForRecord() {
        audioEngine!.inputNode.removeTap(onBus: recordBus)
        let input = audioEngine!.inputNode
        let inputFormat = input.outputFormat(forBus: recordBus)
        let converter = AVAudioConverter(from: inputFormat, to: recordFormat!)!
        let ratio: Float = Float(inputFormat.sampleRate)/Float(recordFormat.sampleRate)
        
        input.installTap(onBus: recordBus, bufferSize: recordBufferSize, format: inputFormat) { (buffer, time) -> Void in
            let inputCallback: AVAudioConverterInputBlock = { inNumPackets, outStatus in
                outStatus.pointee = .haveData
                return buffer
            }
            
            let convertedBuffer = AVAudioPCMBuffer(pcmFormat: self.recordFormat!, frameCapacity: UInt32(Float(buffer.frameCapacity) / ratio))!
            
            var error: NSError?
            let status = converter.convert(to: convertedBuffer, error: &error, withInputFrom: inputCallback)
            assert(status != .error)
            
            if (self.recordFormat?.commonFormat == AVAudioCommonFormat.pcmFormatInt16) {
                let values = self.audioBufferToBytes(convertedBuffer)
                self.sendMicData(values)
            }
        }
    }
    
    private func audioBufferToBytes(_ audioBuffer: AVAudioPCMBuffer) -> [UInt8] {
        let srcLeft = audioBuffer.int16ChannelData![0]
        let bytesPerFrame = audioBuffer.format.streamDescription.pointee.mBytesPerFrame
        let numBytes = Int(bytesPerFrame * audioBuffer.frameLength)
        
        // initialize bytes by 0
        var audioByteArray = [UInt8](repeating: 0, count: numBytes)
        
        srcLeft.withMemoryRebound(to: UInt8.self, capacity: numBytes) { srcByteData in
            audioByteArray.withUnsafeMutableBufferPointer {
                $0.baseAddress!.initialize(from: srcByteData, count: numBytes)
            }
        }
        
        return audioByteArray
    }
    
    private func startRecording(_ result: @escaping FlutterResult) {
        resetEngineForRecord()
        startEngine()
        sendRecorderStatus(SoundStreamStatus.Recording)
        result(true)
    }
    
    private func stopRecording(_ result: @escaping FlutterResult) {
        audioEngine!.inputNode.removeTap(onBus: recordBus)
        sendRecorderStatus(SoundStreamStatus.Stopped)
        result(true)
    }
    
    private func sendMicData(_ data: [UInt8]) {
        let channelData = FlutterStandardTypedData(bytes: NSData(bytes: data, length: data.count) as Data)
        sendEventMethod("dataPeriod", channelData)
    }
    
    private func dispose(_ result: @escaping FlutterResult) {
        audioEngine!.inputNode.removeTap(onBus: recordBus)
        stopEngine()
        result(true)
    }
    
    // #endregion
    
    // #region helpers
    private func printMessage(_ message: String){
        if(debugLogging){
            print(message)
        }
    }
    // #endregion
}
