import 'dart:async';
import 'dart:typed_data';

import 'package:flutter/services.dart';

import '../support/logging/augnito_print_logger.dart';

enum SoundStreamStatus {
  unset,
  initialized,
  recording,
  stopped,
}

/// Bus between high level consumers and plugin native code.
class AugnitoMicStream {
  // channel
  static const _channelName =
      "com.scribetechglobal.augnito_flutter_sdk/methods";

  // methods
  static const _methodInitializeRecorder = "initializeRecorder";
  static const _methodStartRecording = "startRecording";
  static const _methodStopRecording = "stopRecording";
  static const _methodCheckPermission = "checkPermission";
  static const _methodDispose = "dispose";

  // event const
  static const _eventDataPeriod = "dataPeriod";
  static const _eventRecorderStatus = "recorderStatus";

  // param const
  static const _paramSampleRate = "sampleRate";
  static const _paramShowLogs = "showLogs";

  // channel method calls
  static const _channelMethodCallPlatformEvent = "platformEvent";

  // statuses
  static const _eventStatusUnset = "Unset";

  // control
  bool _isRecording = false;
  bool get isRecording => _isRecording;

  // supporting objects
  final _audioStreamController = StreamController<Uint8List>.broadcast();
  final _recorderStatusController =
      StreamController<SoundStreamStatus>.broadcast();
  final _eventsStreamController = StreamController<dynamic>.broadcast();

  // method channel
  static const MethodChannel _channel = MethodChannel(_channelName);

  // logging
  late final bool _enableLogs;
  AugnitoPrintLogger? _logger;

  static AugnitoMicStream? _instance;
  AugnitoMicStream._internal(this._enableLogs) {
    _channel.setMethodCallHandler(_onMethodCall);
    _eventsStreamController.stream.listen(_eventListener);
    _recorderStatusController.add(SoundStreamStatus.unset);
    _audioStreamController.add(Uint8List(0));
    _logger = _enableLogs ? AugnitoPrintLogger() : null;
  }

  factory AugnitoMicStream(bool enableLogs) {
    return _instance ?? AugnitoMicStream._internal(enableLogs);
  }

  /// Initialize Recorder with specified [sampleRate]
  Future<dynamic> initialize(int sampleRate) async {
    try {
      return await _channel.invokeMethod<dynamic>(_methodInitializeRecorder, {
        _paramSampleRate: sampleRate,
        _paramShowLogs: _enableLogs,
      });
    } catch (e) {
      // enable to initialize: check permissions or parameters
      _logger?.log(this, e);
      return false;
    }
  }

  /// Start recording. Recorder will start pushing audio chunks (PCM 16bit data)
  /// to [audiostream] as Uint8List
  Future<dynamic> startRecording() async {
    var result = false;
    try {
      result = await _channel.invokeMethod<dynamic>(_methodStartRecording);
    } catch (e) {
      _logger?.log(this, e);
    }
    _isRecording = result;
    return result;
  }

  /// Recorder will stop recording and sending audio chunks to the [audioStream].
  Future<dynamic> stopRecording() async {
    var result = false;
    try {
      await _channel.invokeMethod<dynamic>(_methodStopRecording);
    } catch (e) {
      _logger?.log(this, e);
    }
    _isRecording = false;
    return result;
  }

  Future<dynamic> checkPermission() async {
    return await _channel.invokeMethod<dynamic>(_methodCheckPermission);
  }

  void dispose() async {
    await stopRecording();
    await _eventsStreamController.close();
    await _recorderStatusController.close();
    await _audioStreamController.close();
    try {
      await _channel.invokeMethod<dynamic>(_methodDispose);
    } catch (e) {
      _logger?.log(this, e);
    }
  }

  /// Stream of PCM 16bit data from Microphone
  Stream<Uint8List> get audioStream => _audioStreamController.stream;

  /// Current status of the [RecorderStream]
  Stream<SoundStreamStatus> get statusStream =>
      _recorderStatusController.stream;

  void _eventListener(dynamic event) {
    if (event == null) return;
    final String eventName = event["name"] ?? "";
    switch (eventName) {
      case _eventDataPeriod:
        final Uint8List audioData = Uint8List.fromList(event["data"]);
        if (audioData.isNotEmpty) _audioStreamController.add(audioData);
        break;
      case _eventRecorderStatus:
        final String status = event["data"] ?? _eventStatusUnset;
        _recorderStatusController.add(SoundStreamStatus.values.firstWhere(
          (value) => _enumToString(value) == status,
          orElse: () => SoundStreamStatus.unset,
        ));
        break;
    }
  }

  Future<dynamic> _onMethodCall(MethodCall call) async {
    if (call.method == _channelMethodCallPlatformEvent) {
      _eventsStreamController.add(call.arguments);
    }
    return null;
  }

  String _enumToString(Object o) => o.toString().split('.').last;
}
