import 'dart:async';

import '../../support/logging/augnito_print_logger.dart';
import '/audio/augnito_mic_stream.dart';

class AugnitoAudioStream {
  static const defaultSampleRate = 16000;

  late final AugnitoMicStream _augnitoMicStream;
  late StreamSubscription? _audioStream;

  bool get isRecording => _augnitoMicStream.isRecording;

  int _sampleRate = AugnitoAudioStream.defaultSampleRate;

  AugnitoPrintLogger? _logger;

  AugnitoAudioStream(
      {sampleRate = AugnitoAudioStream.defaultSampleRate, enableLogs = false}) {
    _augnitoMicStream = AugnitoMicStream(enableLogs);
    _sampleRate = sampleRate;
    _logger = enableLogs ? AugnitoPrintLogger() : null;
  }

  /// Attempts to initialize the audio stream and begins listening to the mic.
  /// It can return false if an error is encountered or if the permission
  /// to record audio has not been granted
  Future<bool> startListening(
      {Function(dynamic)? processSamples, Function(dynamic)? onError}) async {
    final initialized = await _augnitoMicStream.initialize(_sampleRate);
    if (!initialized) {
      _logger?.log(this,
          "Unable to initialize. Check that recording permissions are granted.");
      return false;
    }
    final startedRecording = await _augnitoMicStream.startRecording();
    if (!startedRecording) {
      _logger?.log(this, "unable to start recording.");
      return false;
    }
    _audioStream = _augnitoMicStream.audioStream.listen((data) {
      if (processSamples != null) {
        processSamples(data);
      }
    }, onError: (error) {
      if (onError != null) {
        onError(error);
      }
    });
    return true;
  }

  /// Attempts to close the microphone stream.
  Future<bool> stopListening() async {
    if (_augnitoMicStream.isRecording) {
      _audioStream?.cancel();
      return await _augnitoMicStream.stopRecording();
    }
    return true;
  }

  /// Checks if the permissions for recording are enabled.
  Future<bool> checkPermissions() async {
    return await _augnitoMicStream.checkPermission();
  }

  /// Cancels the audio stream and calls dispose on the native code handler.
  void dispose() async {
    if (_audioStream != null) {
      await _audioStream?.cancel();
      _audioStream = null;
    }
    _augnitoMicStream.dispose();
  }
}
