import 'dart:convert';

import 'package:flutter/foundation.dart';

import '../config/augnito_config.dart';
import '../config/sdk_config.dart';
import 'commands/augnito_command_builder.dart';
import '../support/logging/augnito_print_logger.dart';
import 'core/augnito_audio_stream.dart';
import 'core/augnito_web_client.dart';
import 'dto/augnito_server_response.dart';
import 'dto/meta_event_response.dart';
import 'dto/result_response.dart';
import 'models/action_recipe.dart';
import 'models/speech_to_text_output.dart';
import 'support/augnito_event_types.dart';
import 'support/dictation_error.dart';

/// Manages connection and audio transmission to the server
class DictationManager {
  /// called when the server sends a successful pairing message
  Function()? onConnected;

  /// called when the socket disconnects
  Function()? onDisconnected;

  /// called on errors
  Function(DictationError)? onError;

  /// Hypothesis result
  Function(String hypothesis)? onPartialResult;

  /// Final result
  Function(String transcription)? onFinalResult;

  // Command
  Function(ActionRecipe actionRecipe)? onCommandResult;

  late AugnitoAudioStream _augnitoAudioStream;
  late AugnitoWebClient _augnitoWebClient;

  bool _isBusy = false;
  bool get isBusy => _isBusy;

  AugnitoPrintLogger? _logger;

  /// Creates a Dictation Manager object. Supports callback methods for different events:
  DictationManager(AugnitoConfig augnitoConfig,
      {this.onConnected,
      this.onError,
      this.onDisconnected,
      this.onFinalResult,
      this.onPartialResult,
      this.onCommandResult,
      bool enableLogs = false}) {
    _logger = enableLogs ? AugnitoPrintLogger() : null;
    _augnitoAudioStream = AugnitoAudioStream(enableLogs: enableLogs);
    _augnitoWebClient = AugnitoWebClient(SDKConfig(augnitoConfig).speechUrl,
        onMessage: _socketHandleMessage,
        onDisconnected: _onSocketDisconnected,
        onError: _onSocketError,
        enableLogs: enableLogs);
  }

  DictationManager.fromCustomSpeechURL(String customSpeechURL,
      {this.onConnected,
      this.onError,
      this.onDisconnected,
      this.onFinalResult,
      this.onPartialResult,
      this.onCommandResult,
      bool enableLogs = false}) {
    _logger = enableLogs ? AugnitoPrintLogger() : null;
    _augnitoAudioStream = AugnitoAudioStream(enableLogs: enableLogs);
    _augnitoWebClient = AugnitoWebClient(customSpeechURL,
        onMessage: _socketHandleMessage,
        onDisconnected: _onSocketDisconnected,
        onError: _onSocketError,
        enableLogs: enableLogs);
  }

  /// Initializes or stops the dictation flow.
  Future<void> toggleDictation() async {
    if (_isBusy) {
      return;
    }
    _isBusy = true;
    if (_augnitoAudioStream.isRecording) {
      await _stopDictation();
    } else {
      _startDictation();
    }
  }

  /// Attempts to start a connection with the server. If successful starts the
  /// audio plugin for recording.
  void _startDictation() async {
    if (_augnitoAudioStream.isRecording) {
      return;
    }
    final isConnected = await _augnitoWebClient.open();
    if (!isConnected) {
      _isBusy = false;
      _dispatchError(DictationError(DictationErrorType.socketConnectionError,
          "unable to open the socket. check internet connection."));
    }
  }

  /// Stops the recorder audio capture and closes the socket connection
  Future<void> _stopDictation() async {
    await _augnitoAudioStream.stopListening();
    _augnitoWebClient.close();
  }

  /// Closes the underlying streams and objects. Calls the dispose method
  /// on the native code as well.
  void dispose() async {
    onConnected = null;
    onDisconnected = null;
    onError = null;
    _augnitoWebClient.close();
    await _augnitoAudioStream.stopListening();
    _augnitoAudioStream.dispose();
  }

  //#region callbacks
  Future<void> _onSocketDisconnected() async {
    await _augnitoAudioStream.stopListening();
    _augnitoWebClient.close();
    _isBusy = false;
    _dispatchOnDisconnected();
  }

  Future<void> _onSocketError(dynamic error) async {
    await _augnitoAudioStream.stopListening();
    _augnitoWebClient.close();
    _isBusy = false;
    _dispatchError(DictationError(DictationErrorType.socketDisconnect, ""));
  }
  //#endregion

  void _socketHandleMessage(dynamic message) async {
    try {
      final augnitoServerResponse =
          AugnitoServerResponse.fromJson(jsonDecode(message));
      if (augnitoServerResponse.status != 0) {
        return;
      }
      final metaEvent = augnitoServerResponse.metaEventResponse;
      if (augnitoServerResponse.type == AugnitoServerResponse.typeMeta &&
          metaEvent != null &&
          metaEvent.isNotEmpty()) {
        switch (metaEvent.type) {
          case AugnitoEventTypes.sessionCreated:
            // After successful authenticate, server creates an unique ID for each speech session and sends it back to client app for reference.
            // Client app can store this is it requires.
            await _processSessionCreated(augnitoServerResponse);
            return;
          case AugnitoEventTypes.invalidLoginToken:
            // This event happens when one of following is invalid.
            // AccountCode, AccessKey, Active subscription for trial or paid. lmid.
            await _processInvalidLoginToken(metaEvent);
            return;
          case AugnitoEventTypes.noDictationStopMic:
            // Some time user start mic and forgot after it. start doing discussion with colleague or on phone.
            // In this case mic is on and user is not dictating any valid speech for transcription. Server can detect such situations and send an event to confirm from user.
            _dispatchError(DictationError(
                DictationErrorType.noDictationStopMic, metaEvent.value));
            return;
          case AugnitoEventTypes.templateDictation:
            return;
          case AugnitoEventTypes.serviceDown:
            // Very rare, But This event will come when Speech server's any internal component down.
            await _processServiceDown(metaEvent);
            return;
          case AugnitoEventTypes.lowBandwidth:
            // Speech API need continues upload speed of 32KBps if it raw audio data with 16k sampling rate.
            // If fluctuation in internet than speech output may be delayed. It's good to notify that speech may delayed due to poor network connection.
            // Client app can use this event to show un attendant popup to indicate network status.
            _dispatchError(DictationError(
                DictationErrorType.lowBandwidth, metaEvent.value));
            return;
          default:
            _dispatchError(
                DictationError(DictationErrorType.unknown, metaEvent.value));
            return;
        }
      } else if (augnitoServerResponse.resultResponse != null) {
        await _processResultResponse(augnitoServerResponse.resultResponse!);
      }
    } catch (e) {
      _logger?.log(this, e);
      _dispatchError(DictationError(DictationErrorType.unknown, e.toString()));
    }
  }

  /// Processes a [ResultResponse] from the server.
  /// Will attempt to compute the commands in a separate isolate if the
  /// response contains a final text
  /// Communicates to subscriber via dispatching the corresponding functions.
  _processResultResponse(ResultResponse resultResponse) async {
    final speechToTextOutput =
        await compute<ResultResponse, SpeechToTextOutput>(
            _buildFinalSpeechToTextOutput, resultResponse);

    if (!speechToTextOutput.isFinalText) {
      _dispatchOnPartialResult(speechToTextOutput);
      return;
    }
    if (speechToTextOutput.actionRecipe.isCommand) {
      _dispatchOnCommandResult(speechToTextOutput);
    } else {
      _dispatchOnFinalResult(speechToTextOutput);
    }
  }

  /// To be called when received a service down message.
  _processServiceDown(MetaEventResponse metaEvent) async {
    _dispatchError(
        DictationError(DictationErrorType.serviceDown, metaEvent.value));
    await _stopDictation();
  }

  /// To be called when an invalid token message is received.
  _processInvalidLoginToken(MetaEventResponse metaEvent) async {
    _dispatchError(DictationError(
        DictationErrorType.invalidAuthCredentials, metaEvent.value));
    await _stopDictation();
  }

  /// To be called after a session created response is received.
  /// Attempts to initialize the native audio stream. If unable to start for
  /// any reason it will close the socket and return the corresponding error.
  _processSessionCreated(AugnitoServerResponse augnitoServerResponse) async {
    _dispatchOnConnected();
    final started = await _augnitoAudioStream.startListening(
        processSamples: _processSamples, onError: _onAudioRecorderError);

    if (!started) {
      final hasPermission = await _augnitoAudioStream.checkPermissions();
      if (!hasPermission) {
        _dispatchError(DictationError(
            DictationErrorType.audioRecorderCouldNotInitializePermission,
            "Unable to initialize audio. Record audio permission not granted."));
      } else {
        _dispatchError(DictationError(
            DictationErrorType.audioRecorderCouldNotInitializeGeneric,
            "Unable to initialize audio."));
      }

      await _stopDictation();
      return;
    }

    _isBusy = false;
  }

  /// processes a [ResultResponse] and returns a [SpeechToTextOutput] object.
  /// Intended to be use with compute.
  static SpeechToTextOutput _buildFinalSpeechToTextOutput(
      ResultResponse resultResponse) {
    var actionRecipe = ActionRecipe(receivedText: resultResponse.transcript);
    if (!resultResponse.isFinal) {
      return SpeechToTextOutput(actionRecipe, false);
    } else {
      actionRecipe.action = resultResponse.action;
      actionRecipe.isCommand =
          actionRecipe.isStaticCommand = resultResponse.isCommand;
      AugnitoCommandBuilder.build(actionRecipe);
      return SpeechToTextOutput(actionRecipe, true);
    }
  }

  void _dispatchOnCommandResult(SpeechToTextOutput speechToTextOutput) {
    _logger?.log(this, "Dispatching onCommandResult");
    if (onCommandResult != null) {
      onCommandResult!(speechToTextOutput.actionRecipe);
    }
  }

  void _dispatchOnFinalResult(SpeechToTextOutput speechToTextOutput) {
    _logger?.log(this, "Dispatching onFinalResult");
    if (onFinalResult != null) {
      onFinalResult!(speechToTextOutput.actionRecipe.receivedText);
    }
  }

  void _dispatchOnPartialResult(SpeechToTextOutput speechToTextOutput) {
    _logger?.log(this, "Dispatching onPartialResult");
    if (onPartialResult != null) {
      onPartialResult!(speechToTextOutput.actionRecipe.receivedText);
    }
  }

  void _dispatchOnConnected() {
    _logger?.log(this, "Dispatching onConnected");
    if (onConnected != null) {
      onConnected!();
    }
  }

  void _dispatchOnDisconnected() {
    _logger?.log(this, "Dispatching onDisconnected");
    if (onDisconnected != null) {
      onDisconnected!();
    }
  }

  void _dispatchError(DictationError dictationError) {
    _logger?.log(this,
        "Dispatching error type ${dictationError.errorType}: ${dictationError.errorMessage}");
    if (onError != null) {
      onError!(dictationError);
    }
  }

  void _onAudioRecorderError(dynamic error) {
    _logger?.log(this, "Dispatching recorder error: ${error.toString()}");
    _dispatchError(
        DictationError(DictationErrorType.audioRecorder, error.toString()));
  }

  void _processSamples(dynamic samples) {
    _augnitoWebClient.sendMessage(samples);
  }
}
