import 'package:flutter/material.dart';

abstract class TextFieldProcessor {
  /// Adds content to a widget via the TextEditingController.
  static addContent(
      String receivedText, TextEditingController textEditingController) {
    final selectionStart = textEditingController.selection.start;
    final editorReadyText = _prepareTextForEditor(
        receivedText, selectionStart, textEditingController.text);
    _writeAtCaret(textEditingController.text, editorReadyText, selectionStart,
        textEditingController);
  }

  /// Attempts to select the last N lines starting from the current caret position.
  static void selectLastLines(TextEditingController textEditingController,
      {int offset = 1}) {
    if (offset < 1) {
      return;
    }
    final caretPosition = textEditingController.selection.start;
    final rightToCaret = textEditingController.text.substring(caretPosition);

    final textUntilNextDelimiter =
        textEditingController.text.substring(0, caretPosition) +
            rightToCaret.split("\n").first;

    final selectedLines =
        textUntilNextDelimiter.split("\n").reversed.take(offset);

    final int totalCharacters =
        selectedLines.join().length + selectedLines.length - 1;

    textEditingController.selection = TextSelection(
        baseOffset: textUntilNextDelimiter.length - totalCharacters,
        extentOffset: textUntilNextDelimiter.length);
  }

  /// Attempts to select the last N words counting from the caret position.
  static void selectLastWords(TextEditingController textEditingController,
      {int offset = 1}) {
    if (offset < 1) {
      return;
    }
    var caretPosition = textEditingController.selection.start;
    var rightToCaret = textEditingController.text.substring(caretPosition);
    final textUntilNextDelimiter =
        textEditingController.text.substring(0, caretPosition) +
            rightToCaret.split(" ").first;

    // update the caret position so its positioned at the end of the obtained text

    caretPosition = textUntilNextDelimiter.length;

    final lastNWordsRegExp =
        RegExp(r"\S+(?:\s+\S+){0," + (offset - 1).toString() + r"}$");

    final match = lastNWordsRegExp.firstMatch(textUntilNextDelimiter.trim());
    if (match != null) {
      final selectedText = match.group(0);

      final selectionStartOffset = caretPosition - selectedText!.length;
      textEditingController.selection = TextSelection(
          baseOffset: selectionStartOffset, extentOffset: caretPosition);
    }
  }

  /// Deletes the text contents within a selection given a TextEditingController.
  static void deleteSelection(TextEditingController textEditingController) {
    if (textEditingController.text.isEmpty ||
        textEditingController.selection.isCollapsed) {
      return;
    }

    final selectedText = textEditingController.text.substring(
        textEditingController.selection.start,
        textEditingController.selection.end);

    final newCaretPosition =
        textEditingController.selection.end - selectedText.length;

    textEditingController.text = textEditingController.text
            .substring(0, textEditingController.selection.start) +
        textEditingController.text
            .substring(textEditingController.selection.end);

    textEditingController.selection =
        TextSelection.collapsed(offset: newCaretPosition);
  }

  static void _writeAtCaret(String currentText, String hypothesisText,
      int caretPosition, TextEditingController textEditingController) {
    final leftText = currentText.substring(0, caretPosition);
    final rightText = currentText.substring(caretPosition);
    textEditingController.text = "$leftText$hypothesisText$rightText";
    final finalCaretPosition = leftText.length + hypothesisText.length;
    textEditingController.selection =
        TextSelection.collapsed(offset: finalCaretPosition);
  }

  static String _prepareTextForEditor(
      String receivedText, int caretPosition, String currentText) {
    var editorReady = receivedText;
    if (currentText.isEmpty) {
      editorReady = _trimLeft(editorReady);
      editorReady = _capitalize(editorReady);
    }

    // Remove prefix spaces for newline and new paragraph
    if (editorReady == " \n" || editorReady == " \n\n") {
      editorReady = editorReady.trimLeft();
    }

    if (currentText.length > 1 &&
        currentText[caretPosition - 1] == " " &&
        currentText[caretPosition - 2] == ".") {
      editorReady = _trimLeft(editorReady);
      editorReady = _capitalize(editorReady);
    }

    // If already segment has space then remove server given space, as the user has moved the cursor.
    if (currentText.isNotEmpty && currentText[caretPosition - 1] == " ") {
      editorReady = _trimLeft(editorReady);
    }

    // Start of new line
    if (currentText.isNotEmpty && currentText[caretPosition - 1] == "\n") {
      editorReady = _trimLeft(editorReady);
      editorReady = _capitalize(editorReady);
    }

    final delimiters = [".", ":", "#"];
    if (currentText.isNotEmpty &&
        delimiters.contains(currentText[caretPosition - 1])) {
      var trimmedLeft = _trimLeft(editorReady);
      if (trimmedLeft.isNotEmpty) {
        editorReady = " ${trimmedLeft[0].toUpperCase()}";
      }
      if (trimmedLeft.length > 1) {
        editorReady += trimmedLeft.substring(1);
      }
    }

    if (editorReady.trim() == "." || editorReady.trim() == ":") {
      editorReady = _trimLeft(editorReady);
    }

    return editorReady;
  }

  static String _trimLeft(String receivedText) {
    const newLocal = "@newline@";
    receivedText = receivedText
        .replaceAllMapped(RegExp(r'\n', caseSensitive: false), (match) {
      return newLocal;
    });
    receivedText = receivedText.trimLeft();
    receivedText = receivedText.replaceAll(RegExp(newLocal), "\n");
    return receivedText;
  }

  static _capitalize(String input) =>
      input[0].toUpperCase() + input.substring(1);
}
